(function() {
  var Filter, Query, Serializer, each, map;

  each = require('lodash/each');

  map = require('lodash/map');

  Serializer = require('./serializer');

  ({Filter} = require('firebase-admin/firestore'));

  Query = (function() {
    var andClause, filters, ops, orClause;

    class Query {
      constructor(q = [], serializer = null) {
        this.serializer = serializer != null ? serializer : new Serializer();
        this.q = q;
        return;
      }

      prepare(col_ref, serializer = null) {
        var fbq, s;
        s = serializer != null ? serializer : this.serializer;
        fbq = col_ref;
        each(this.q, function(statement) {
          fbq = ops[statement[0]](fbq, s, ...statement.slice(1));
        });
        return fbq;
      }

    };

    Query.prototype.type = 'Query';

    // fbq = firestore query, s = serializer, f = field, v(s) = value(s), cl(s) = clause(s)
    filters = {
      'eq': function(s, f, v) {
        return Filter.where(s.fmt[s.fmts.field.db](f), '==', v);
      },
      'eq-any': function(s, f, vs) {
        return Filter.where(s.fmt[s.fmts.field.db](f), 'in', vs);
      },
      'gt': function(s, f, v) {
        return Filter.where(s.fmt[s.fmts.field.db](f), '>', v);
      },
      'gte': function(s, f, v) {
        return Filter.where(s.fmt[s.fmts.field.db](f), '>=', v);
      },
      'includes': function(s, f, v) {
        return Filter.where(s.fmt[s.fmts.field.db](f), 'array-contains', v);
      },
      'includes-any': function(s, f, vs) {
        return Filter.where(s.fmt[s.fmts.field.db](f), 'array-contains-any', vs);
      },
      'lt': function(s, f, v) {
        return Filter.where(s.fmt[s.fmts.field.db](f), '<', v);
      },
      'lte': function(s, f, v) {
        return Filter.where(s.fmt[s.fmts.field.db](f), '<=', v);
      },
      'not-eq': function(s, f, v) {
        return Filter.where(s.fmt[s.fmts.field.db](f), '!=', v);
      },
      'not-eq-any': function(s, f, vs) {
        return Filter.where(s.fmt[s.fmts.field.db](f), 'not-in', vs);
      }
    };

    ops = {
      'and': function(fbq, s, ...cls) {
        return fbq.where(andClause(cls, s));
      },
      'eq': function(fbq, s, f, v) {
        return fbq.where(s.fmt[s.fmts.field.db](f), '==', v);
      },
      'eq-any': function(fbq, s, f, vs) {
        return fbq.where(s.fmt[s.fmts.field.db](f), 'in', vs);
      },
      'gt': function(fbq, s, f, v) {
        return fbq.where(s.fmt[s.fmts.field.db](f), '>', v);
      },
      'gte': function(fbq, s, f, v) {
        return fbq.where(s.fmt[s.fmts.field.db](f), '>=', v);
      },
      'includes': function(fbq, s, f, v) {
        return fbq.where(s.fmt[s.fmts.field.db](f), 'array-contains', v);
      },
      'includes-any': function(fbq, s, f, vs) {
        return fbq.where(s.fmt[s.fmts.field.db](f), 'array-contains-any', vs);
      },
      'limit': function(fbq, _, n) {
        return fbq.limit(n);
      },
      'lt': function(fbq, s, f, v) {
        return fbq.where(s.fmt[s.fmts.field.db](f), '<', v);
      },
      'lte': function(fbq, s, f, v) {
        return fbq.where(s.fmt[s.fmts.field.db](f), '<=', v);
      },
      'not-eq': function(fbq, s, f, v) {
        return fbq.where(s.fmt[s.fmts.field.db](f), '!=', v);
      },
      'not-eq-any': function(fbq, s, f, vs) {
        return fbq.where(s.fmt[s.fmts.field.db](f), 'not-in', vs);
      },
      'or': function(fbq, s, ...cls) {
        return fbq.where(orClause(cls, s));
      },
      'order-by': function(fbq, s, f, d) {
        return fbq.orderBy(s.fmt[s.fmts.field.db](f), d != null ? d : 'asc');
      },
      'start-at': function(fbq, _, v) {
        return fbq.startAt(...v);
      }
    };

    orClause = function(cls = [], serializer = null) {
      var s;
      s = serializer != null ? serializer : this.serializer;
      return Filter.or(...(map(cls, (function(cl) {
        return filters[cl[0]](s, ...cl.slice(1));
      }))));
    };

    andClause = function(cls = [], serializer = null) {
      var s;
      s = serializer != null ? serializer : this.serializer;
      return Filter.and(...(map(cls, (function(cl) {
        return filters[cl[0]](s, ...cl.slice(1));
      }))));
    };

    return Query;

  }).call(this);

  module.exports = Query;

}).call(this);
